#!/usr/bin/env python3

import os
import re
import subprocess
from netaddr import IPNetwork
from sysrepo_api import sysrepo_api3 as api
import vppapi_if.vppapi_if as vppif
import ipaddress

file_list_tacacs = '/etc/pam.d/pam_tacacs /etc/pam.d/sshd /etc/pam.d/login'
pam_radius_auth_fallback_disable = "auth required /lib/x86_64-linux-gnu/security/pam_radius_auth.so\n"
pam_radius_auth_fallback_enable = "auth required /lib/x86_64-linux-gnu/security/pam_radius_auth.so fallback\n"
pam_radius_fallback_token_disable = "auth required /lib/x86_64-linux-gnu/security/pam_radius_auth.so config_file=/etc/pam_radius_auth.conf token\n"
pam_radius_fallback_token_enable = "auth required /lib/x86_64-linux-gnu/security/pam_radius_auth.so config_file=/etc/pam_radius_auth.conf token fallback\n"
pam_radius_files = ['/etc/pam.d/sshd', '/etc/pam.d/login']
sshd_file = '/etc/pam.d/sshd'
login_file = '/etc/pam.d/login'
IPTABLES = {'/sbin/iptables', '/sbin/ip6tables'}
chain_name = "cli-whitelist"
proto = "tcp"
ssh_port_file = '/config/ssh_port'

def get_ssh_port():
    ssh_port = 22222
    try:
        with open(ssh_port_file, 'r') as file:
            content = file.read().strip()
            ssh_port = int(content) if content else 22222

    except FileNotFoundError:
        print(f"Error: The file {ssh_port_file} was not found. Using default port 22222.")
        ssh_port = 22222
    except Exception as e:
        print(f"Error occurred: {e} while getting the ssh-port.")
        ssh_port = 22222

    return ssh_port

dst_port = get_ssh_port()


class SysrepoSecurity(api.SysrepoAPI, object):
    def __init__(self,
                 app_name="security",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/security:security",
                 indexes={"access-source-ip": "address"},
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority)

    def post_apply_config(self, change_set):
        if not change_set:
            return

        super(SysrepoSecurity, self).post_apply_config(change_set)

    def get_interfaces(self, startup=False):
        config_indexed = {}
        sys_conf = self.get_startup_config(full_config=True) if startup else self.get_running_config(full_config=True)
        for interface in sys_conf['config']['interface']:
            config_indexed[interface['name']] = interface
        return config_indexed

    def apply_config(self, change_set=None, apply_startup=False):
        if (apply_startup):
            self._apply_startup_config()
            return True

        # get old values and save them in dictionary
        radius_dict = {"server": "", "port": "", "key": "", "timeout": ""}
        backup_radius_dict = {"backup-server": "", "backup-port": "", "backup-key": "", "backup-timeout": ""}

        with open('/etc/pam_radius_auth.conf', 'r') as file:
            f = file.readlines()

            # first server
            server_old = f[0].split(":")
            if len(server_old) > 5:
                server_params = server_old[-1:]
                server_old.pop()
                server_old = ':'.join(server_old)
                port_old = server_params[0].split(' ')[0]
                key_old = server_params[0].split(' ')[1]
                timeout_old = server_params[0].split(' ')[2].strip()
            else:
                server_old = f[0].split(':')[0].strip()
                port_old_split = f[0].split(':')[1]
                port_old = port_old_split.split(' ')[0]
                key_old = f[0].split(' ')[1]
                timeout_old = f[0].split(' ')[2].strip()

            # second server
            server_backup_old = f[1].split(':')
            if len(server_backup_old) > 5:
                server_params = server_backup_old[-1:]
                server_backup_old.pop()
                server_backup_old = ':'.join(server_backup_old)
                port_backup_old = server_params[0].split(' ')[0]
                key_backup_old = server_params[0].split(' ')[1]
                timeout_backup_old = server_params[0].split(' ')[2].strip()
            else:
                server_backup_old = f[1].split(':')[0].strip()
                port_backup_old_split = f[1].split(':')[1]
                port_backup_old = port_backup_old_split.split(' ')[0]
                key_backup_old = f[1].split(' ')[1]
                timeout_backup_old = f[1].split(' ')[2].strip()

        radius_dict["server"] = server_old
        radius_dict["port"] = port_old
        radius_dict["key"] = key_old
        radius_dict["timeout"] = timeout_old

        backup_radius_dict["backup-server"] = server_backup_old
        backup_radius_dict["backup-port"] = port_backup_old
        backup_radius_dict["backup-key"] = key_backup_old
        backup_radius_dict["backup-timeout"] = timeout_backup_old

        for i in range(0, len(change_set)):
            if not change_set[i]["xpath"].startswith(self.namespace):
                continue

            v_set = change_set[i]
            if v_set["operation"] is "MODIFIED" or v_set["operation"] is "CREATED":
                if v_set["xpath"].find("authentication/auth-type") != -1:
                    self._apply_auth_type(v_set["new_value"])

                elif v_set["xpath"].find("authentication/tacacs-server") != -1:
                    self._apply_tacacs_server(v_set["new_value"])
                elif v_set["xpath"].find("authentication/tacacs-key") != -1:
                    self._apply_tacacs_key(v_set["new_value"])

                elif v_set["xpath"].find("authentication/radius-server") != -1:
                    self._apply_radius_server(v_set["new_value"])
                    radius_dict["server"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/radius-port") != -1:
                    self._apply_radius_port(v_set["new_value"])
                    radius_dict["port"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/radius-key") != -1:
                    self._apply_radius_key(v_set["new_value"])
                    radius_dict["key"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/radius-timeout") != -1:
                    self._apply_radius_timeout(v_set["new_value"])
                    if int(v_set["new_value"]) > 20:
                        radius_dict["timeout"] = "20"
                    elif int(v_set["new_value"]) < 3:
                        radius_dict["timeout"] = "3"
                    else:
                        radius_dict["timeout"] = v_set["new_value"]

                elif v_set["xpath"].find("authentication/fallback-auth") != -1:
                    self._apply_fallback_auth(v_set["new_value"])

                elif v_set["xpath"].find("authentication/backup-radius-server") != -1:
                    backup_radius_dict["backup-server"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/backup-radius-port") != -1:
                    backup_radius_dict["backup-port"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/backup-radius-key") != -1:
                    backup_radius_dict["backup-key"] = v_set["new_value"]
                elif v_set["xpath"].find("authentication/backup-radius-timeout") != -1:
                    if int(v_set["new_value"]) > 20:
                        backup_radius_dict["backup-timeout"] = "20"
                    elif int(v_set["new_value"]) < 3:
                        backup_radius_dict["backup-timeout"] = "3"
                    else:
                        backup_radius_dict["backup-timeout"] = v_set["new_value"]

                elif v_set["xpath"].find("cli-whitelist/enabled") != -1:
                    self._apply_cli_whitelist_enabled(v_set["new_value"])

                elif v_set["xpath"].find("cli-whitelist/access-source-ip") != -1 and v_set["xpath"].find("/address") != -1:
                    if IPNetwork(v_set["new_value"]).version is 4:
                        self._apply_access_source_ip(v_set["new_value"], "/sbin/iptables", dst_port=dst_port)
                    else:
                        self._apply_access_source_ip(v_set["new_value"], "/sbin/ip6tables", dst_port=dst_port)
                elif v_set["xpath"].find("ssh-server/listen-port") != -1:
                    c_ifs = self.get_interfaces()
                    self._apply_ssh_port(v_set["new_value"])
                    self._change_ssh_port_interfaces(c_ifs, v_set['old_value'], v_set['new_value'])
                    if any(iface.get('ip-passthrough', False) for iface in c_ifs.values()):
                        self._reset_ip_pass_rules(c_ifs, v_set["new_value"])
                    if self.get_cli_whitelist_enabled():
                        self._reinitialize_cli_whitelist(self.get_cli_whitelist_config(), v_set["new_value"])
                        print("The whitelist is updated.")
                    else:
                        print("The whitelist is not enabled")

            elif v_set["operation"] is "DELETED":
                if v_set["xpath"].find("authentication/auth-type") != -1:
                    self._apply_auth_type("local")

                elif v_set["xpath"].find("cli-whitelist/access-source-ip") != -1 and v_set["xpath"].find(
                        "/address") != -1:
                    self._remove_access_source_ip('/sbin/iptables', chain_name)

        if radius_dict["server"] != "" or radius_dict["port"] != "" or radius_dict["key"] != "" or radius_dict["timeout"] != "":
            self._write_to_file(radius_dict["server"], radius_dict["port"], radius_dict["key"], radius_dict["timeout"])

        if backup_radius_dict["backup-server"] != "" or backup_radius_dict["backup-port"] != "" or backup_radius_dict["backup-key"] != "" or backup_radius_dict["backup-timeout"] != "":
            self._write_to_file_backup(backup_radius_dict["backup-server"], backup_radius_dict["backup-port"], backup_radius_dict["backup-key"], backup_radius_dict["backup-timeout"])

        return True

    def set_authentication_auth_type(self, auth_type="local"):
        self.set_item("{}/authentication/auth-type".format(self.namespace), auth_type)

    def get_authentication_auth_type(self):
        return self.get_item("{}/authentication/auth-type".format(self.namespace))

    def unset_authentication_auth_type(self):
        self.delete_item("{}/authentication/auth-type".format(self.namespace))

    def set_authentication_tacacs_key(self, tacacs_key):
        self.set_item("{}/authentication/tacacs-key".format(self.namespace), tacacs_key)

    def get_authentication_tacacs_key(self):
        return self.get_item("{}/authentication/tacacs-key".format(self.namespace))

    def unset_authentication_tacacs_key(self):
        self.delete_item("{}/authentication/tacacs-key".format(self.namespace))

    def set_authentication_tacacs_server(self, tacacs_server):
        self.set_item("{}/authentication/tacacs-server".format(self.namespace), tacacs_server)

    def get_authentication_tacacs_server(self):
        return self.get_item("{}/authentication/tacacs-server".format(self.namespace))

    def unset_authentication_tacacs_server(self):
        self.delete_item("{}/authentication/tacacs-server".format(self.namespace))

    def set_authentication_radius_server(self, radius_server):
        self.set_item("{}/authentication/radius-server".format(self.namespace), radius_server)

    def get_authentication_radius_server(self):
        return self.get_item("{}/authentication/radius-server".format(self.namespace))

    def unset_authentication_radius_server(self):
        self.delete_item("{}/authentication/radius-server".format(self.namespace))

    def set_authentication_radius_port(self, radius_port):
        self.set_item("{}/authentication/radius-port".format(self.namespace), radius_port)

    def get_authentication_radius_port(self):
        return self.get_item("{}/authentication/radius-port".format(self.namespace))

    def unset_authentication_radius_port(self):
        self.delete_item("{}/authentication/radius-port".format(self.namespace))

    def set_authentication_radius_key(self, radius_key):
        self.set_item("{}/authentication/radius-key".format(self.namespace), radius_key)

    def get_authentication_radius_key(self):
        return self.get_item("{}/authentication/radius-key".format(self.namespace))

    def unset_authentication_radius_key(self):
        self.delete_item("{}/authentication/radius-key".format(self.namespace))

    def set_authentication_radius_timeout(self, radius_timeout):
        self.set_item("{}/authentication/radius-timeout".format(self.namespace), radius_timeout)

    def get_authentication_radius_timeout(self):
        return self.get_item("{}/authentication/radius-timeout".format(self.namespace))

    def unset_authentication_radius_timeout(self):
        self.delete_item("{}/authentication/radius-timeout".format(self.namespace))

    def set_fallback_authentication(self, fallback_auth):
        self.set_item("{}/authentication/fallback-auth".format(self.namespace), fallback_auth)

    def get_fallback_authentication(self):
        return self.get_item("{}/authentication/fallback-auth".format(self.namespace))

    def unset_fallback_authentication(self):
        self.delete_item("{}/authentication/fallback-auth".format(self.namespace))

    def set_authentication_backup_radius_server(self, backup_radius_server):
        self.set_item("{}/authentication/backup-radius-server".format(self.namespace), backup_radius_server)

    def get_authentication_backup_radius_server(self):
        return self.get_item("{}/authentication/backup-radius-server".format(self.namespace))

    def unset_authentication_backup_radius_server(self):
        self.delete_item("{}/authentication/backup-radius-server".format(self.namespace))

    def set_authentication_backup_radius_port(self, backup_radius_port):
        self.set_item("{}/authentication/backup-radius-port".format(self.namespace), backup_radius_port)

    def get_authentication_backup_radius_port(self):
        return self.get_item("{}/authentication/backup-radius-port".format(self.namespace))

    def unset_authentication_backup_radius_port(self):
        self.delete_item("{}/authentication/backup-radius-port".format(self.namespace))

    def set_authentication_backup_radius_key(self, backup_radius_key):
        self.set_item("{}/authentication/backup-radius-key".format(self.namespace), backup_radius_key)

    def get_authentication_backup_radius_key(self):
        return self.get_item("{}/authentication/backup-radius-key".format(self.namespace))

    def unset_authentication_backup_radius_key(self):
        self.delete_item("{}/authentication/backup-radius-key".format(self.namespace))

    def set_authentication_backup_radius_timeout(self, backup_radius_timeout):
        self.set_item("{}/authentication/backup-radius-timeout".format(self.namespace), backup_radius_timeout)

    def get_authentication_backup_radius_timeout(self):
        return self.get_item("{}/authentication/backup-radius-timeout".format(self.namespace))

    def unset_authentication_backup_radius_timeout(self):
        self.delete_item("{}/authentication/backup-radius-timeout".format(self.namespace))

    def set_cli_whitelist_enabled(self, enabled=False):
        self.set_item("{}/cli-whitelist/enabled".format(self.namespace), enabled)

    def get_cli_whitelist_enabled(self):
        return self.get_item("{}/cli-whitelist/enabled".format(self.namespace))

    def unset_cli_whitelist_enabled(self):
        self.delete_item("{}/cli-whitelist/enabled".format(self.namespace))

    def set_cli_whitelist_access_source_ip(self, address):
        self.set_item("{}/cli-whitelist/access-source-ip[address='{}']".format(self.namespace, address), address)

    def get_cli_whitelist_access_source_ip(self, address):
        return self.get_item("{}/cli-whitelist/access-source-ip[address='{}']".format(self.namespace, address))

    def unset_cli_whitelist_access_source_ip(self, address):
        self.delete_item("{}/cli-whitelist/access-source-ip[address='{}']".format(self.namespace, address))

    def get_cli_whitelist_config(self):
        enabled = self.get_cli_whitelist_enabled()

        # Get the access-source-ip information
        access_source_ip_info = self.get_cli_whitelist_access_source_ip()

        try:
            address = list(access_source_ip_info.values())[0]

            combined_info = {
                'access-source-ip': [address],
                'enabled': enabled
            }
        except Exception as e:
            self.logger.debug(f"Can not return whitelist config due to {e}")
            return

        return combined_info

    def get_cli_ssh_server_listen_port(self):
        return self.get_item("{}/ssh-server/listen-port".format(self.namespace))

    def set_cli_ssh_server_listen_port(self, port):
        self.set_item("{}/ssh-server/listen-port".format(self.namespace), port)

    def unset_cli_ssh_server_listen_port(self, port):
        self.delete_item("{}/ssh-server/listen-port".format(self.namespace), port)

    def get_cli_whitelist_access_source_ip(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'cli-whitelist' not in config or 'access-source-ip' not in config['cli-whitelist']:
            return config_indexed
        for src_ip in config['cli-whitelist']['access-source-ip']:
            config_indexed[src_ip['address']] = src_ip
        return config_indexed

    def _if_exists_chain(self, iptables):
        cmd = '{} -S {}'.format(iptables, chain_name)
        process = subprocess.Popen(
            str(cmd).split(' '),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        (stdout, stderr) = process.communicate()
        if process.returncode != 0:
            return False

        return True

    def _if_whitelist_enabled(self, iptables):
        cmd = '{} -S INPUT | grep {}'.format(iptables, chain_name)
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True)
        (stdout, stderr) = process.communicate()
        if process.returncode != 0:
            return False

        if not stdout:
            return False

        return True

    def _if_exists_deny_any(self, iptables):
        cmd = '{} -S {} | grep " -j REJECT --reject-with tcp-reset"'.format(iptables, chain_name)
        r = self.execute_shell_command(cmd)

        if not r:
            return False

        return True

    def _create_chain(self, iptables):
        if not self._if_exists_chain(iptables):
            cmd = '{} -N {}'.format(iptables, chain_name)
            self.execute_shell_command(cmd)
            self.logger.info("CLI whitelist chain is created")

        self._add_deny_any_rule(iptables, dst_port)

    def _flush_and_delete_chain(self, iptables, chain_name):
        cmd = '{} -F {}'.format(iptables, chain_name)
        self.execute_shell_command(cmd)
        # cmd = '{} -n -L {}'.format(iptables, chain_name)
        # reference_exist = self.execute_shell_command(cmd)
        # if reference_exist:
        self._disable_cli_whitelist(iptables)
        cmd = '{} -X {}'.format(iptables, chain_name)
        self.execute_shell_command(cmd)

    def _add_deny_any_rule(self, iptables, dst_port):
        # check if exists
        if not self._if_exists_deny_any(iptables):
            if '/sbin/iptables' == iptables:
                cmd = '{} -I {} -p {} --dport {} -s 127.0.0.0/8 -j ACCEPT'.format(iptables, chain_name, proto, dst_port)
                self.execute_shell_command(cmd)
            cmd = '{} -A {} -p {} --dport {} -j REJECT --reject-with tcp-reset'.format(iptables, chain_name, proto, dst_port)
            self.execute_shell_command(cmd)
            self.logger.debug("Deny any rule is added to CLI whitelist chain")

    def _add_rule(self, ip, iptables, dst_port):
        cmd = '{} -I {} -p {} --dport {} -s {} -j ACCEPT'.format(iptables, chain_name, proto, dst_port, ip)
        self.execute_shell_command(cmd)
        self.logger.info("Add host {} to access source addresses".format(ip))

    def _remove_rule(self, ip, iptables):
        cmd = '{} -D {} -p {} --dport {} -s {} -j ACCEPT'.format(iptables, chain_name, proto, dst_port, ip)
        self.execute_shell_command(cmd)
        self.logger.info("Remove host {} from access source addresses".format(ip))

    def _enable_cli_whitelist(self, iptables):
        self._create_chain(iptables)
        if not self._if_whitelist_enabled(iptables):
            cmd = '{} -t filter -A INPUT -j {}'.format(iptables, chain_name)
            self.execute_shell_command(cmd)
            self.logger.info("CLI Whitelist feature is enabled")

    def _disable_cli_whitelist(self, iptables):
        if self._if_whitelist_enabled(iptables):
            cmd = '{} -t filter -D INPUT -j {}'.format(iptables, chain_name)
            self.execute_shell_command(cmd)
            self.logger.info("CLI Whitelist feature is disabled")

    def _restart_apache_service(self):
        self._restart_services("apache2")
        os.system('systemctl restart apache2.service')

    def _restart_ssh_service(self):
        self._restart_services("ssh.service")
        self.execute_shell_command('systemctl --no-block restart ssh.service 2>/dev/null || true')

    # Apply functions
    def _apply_auth_type(self, auth_type):
        if not auth_type:
            raise Exception("Authentication type not set !")

        if auth_type == 'tacacs':
            cmd = 'sed -i "s/^#@include pam_tacacs/@include pam_tacacs/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#auth sufficient pam_tacplus.so/auth sufficient pam_tacplus.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#auth sufficient pam_script.so/auth sufficient pam_script.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#account optional pam_tacplus.so/account optional pam_tacplus.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#account optional pam_script.so/account optional pam_script.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#session optional pam_tacplus.so/session optional pam_tacplus.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#session optional pam_script.so/session optional pam_script.so/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#auth sufficient/auth sufficient/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#session optional/session optional/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^#account optional/account optional/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        else:
            cmd = 'sed -i "s/^@include pam_tacacs/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^auth sufficient pam_tacplus.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^auth sufficient pam_script.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^account optional pam_tacplus.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^account optional pam_script.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^session optional pam_tacplus.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^session optional pam_script.so/#&/" %s' % (sshd_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^auth sufficient/#&/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^session optional/#&/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            cmd = 'sed -i "s/^account optional/#&/" %s' % (login_file)
            subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

        if auth_type == 'tacacs' or auth_type == 'local':
            for file in pam_radius_files:
                with open(file, 'r') as f:
                    for content in f.readlines():
                        if content == pam_radius_auth_fallback_enable or content == pam_radius_auth_fallback_disable:
                            with open(file, 'r') as f:
                                data = f.read().splitlines(True)
                            with open(file, 'w') as f:
                                f.writelines(data[1:])

            self._replace_radius_data("/etc/pam.d/pam_radius_auth", pam_radius_fallback_token_enable, pam_radius_fallback_token_disable)
            self._replace_radius_data('/etc/pam.d/sshd', '\n#@include common-auth', '\n@include common-auth')
            self._replace_radius_data('/etc/pam.d/sshd', '\n#@include common-session', '\n@include common-session')
            self._replace_radius_data('/etc/pam.d/sshd', '\n#@include common-password', '\n@include common-password')

            self._restart_apache_service()

        if auth_type == 'radius':
            for file in pam_radius_files:
                with open(file, 'r') as f:
                    contents = f.readlines()
                    if contents[0] == pam_radius_auth_fallback_enable or contents[0] == pam_radius_auth_fallback_disable:
                        pass
                    else:
                        contents.insert(0, pam_radius_auth_fallback_enable)
                        self._replace_radius_data("/etc/pam.d/pam_radius_auth", pam_radius_fallback_token_disable, pam_radius_fallback_token_enable)
                        with open(file, 'w') as f:
                            contents = "".join(contents)
                            f.write(contents)

            self._replace_radius_data('/etc/pam.d/sshd', '\n@include common-auth', '\n#@include common-auth')
            self._replace_radius_data('/etc/pam.d/sshd', '\n@include common-session', '\n#@include common-session')
            self._replace_radius_data('/etc/pam.d/sshd', '\n@include common-password', '\n#@include common-password')

        self._restart_apache_service()
        self.logger.info("Authentication type %s is applied" % auth_type)

    def _apply_tacacs_server(self, tac_server):
        if not tac_server:
            self.logger.info("TACACS server not set !")
            return

        cmd = 'gawk -i inplace -v ip=%s \'{gsub(/server=.*/,"server="ip,$4); print $0}\' %s' % (tac_server, file_list_tacacs)
        subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        self.logger.info("TACACS server %s is applied " % tac_server)

    def _apply_tacacs_key(self, tac_key):
        if not tac_key:
            self.logger.info("TACACS key not set !")
            return

        cmd = 'gawk -i inplace -v key=%s \'{gsub(/secret=.*/,"secret="key,$5); print $0}\' %s' % (tac_key, file_list_tacacs)
        subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        self.logger.info("TACACS key %s is applied" % tac_key)

    # function to write server_ip, port and key for radius auth to file
    @staticmethod
    def _write_to_file(rad_server, rad_port, rad_key, rad_timeout):

        with open('/etc/pam_radius_auth.conf', 'r') as file:
            f = file.readlines()

            server_backup_old = f[1].split(':')
            if len(server_backup_old) > 5:
                server_params = server_backup_old[-1:]
                server_backup_old.pop()
                server_backup_old = ':'.join(server_backup_old)
                port_backup_old = server_params[0].split(' ')[0]
                key_backup_old = server_params[0].split(' ')[1]
                timeout_backup_old = server_params[0].split(' ')[2].strip()
            else:
                server_backup_old = f[1].split(':')[0].strip()
                port_backup_old_split = f[1].split(':')[1]
                port_backup_old = port_backup_old_split.split(' ')[0]
                key_backup_old = f[1].split(' ')[1]
                timeout_backup_old = f[1].split(' ')[2].strip()

        with open('/etc/pam_radius_auth.conf', 'w+') as file:
            file.write(f"{rad_server}:{rad_port} {rad_key} {rad_timeout.strip()}\n{server_backup_old}:{port_backup_old} {key_backup_old} {timeout_backup_old}")
            file.close()

    @staticmethod
    def _write_to_file_backup(server_backup_old, port_backup_old, key_backup_old, timeout_backup_old):

        with open('/etc/pam_radius_auth.conf', 'r') as file:
            f = file.readlines()

            server_old = f[0].split(":")
            if len(server_old) > 5:
                server_params = server_old[-1:]
                server_old.pop()
                server_old = ':'.join(server_old)
                port_old = server_params[0].split(' ')[0]
                key_old = server_params[0].split(' ')[1]
                timeout_old = server_params[0].split(' ')[2].strip()
            else:
                server_old = f[0].split(':')[0].strip()
                port_old_split = f[0].split(':')[1]
                port_old = port_old_split.split(' ')[0]
                key_old = f[0].split(' ')[1]
                timeout_old = f[0].split(' ')[2].strip()

        with open('/etc/pam_radius_auth.conf', 'w+') as file:
            file.write(f"{server_old}:{port_old} {key_old} {timeout_old}\n{server_backup_old}:{port_backup_old} {key_backup_old} {timeout_backup_old.strip()}")
            file.close()

    @staticmethod
    def _replace_radius_data(file, data_to_replace, replaced_data):
        fin = open(file, "rt")
        data = fin.read()
        data = data.replace(data_to_replace, replaced_data)
        fin.close()
        fin = open(file, "wt")
        fin.write(data)
        fin.close()

    def _apply_radius_server(self, rad_server):
        if not rad_server:
            self.logger.info("RADIUS server not set !")
            return

        self.logger.info("RADIUS server %s is applied " % rad_server)

    def _apply_radius_port(self, rad_port):
        if not rad_port:
            self.logger.info("RADIUS port not set !")
            return

        self.logger.info("RADIUS port %s is applied" % rad_port)

    def _apply_radius_key(self, rad_key):
        if not rad_key:
            self.logger.info("RADIUS key not set !")
            return

        self.logger.info("RADIUS key %s is applied" % rad_key)

    def _apply_radius_timeout(self, rad_timeout):
        if not rad_timeout:
            self.logger.info("RADIUS timeout not set !")
            return

        self.logger.info("RADIUS timeout %s is applied" % rad_timeout)

    def _apply_fallback_auth(self, fallback_auth):
        if not fallback_auth:
            self.logger.info("Fallback authentication is not set!")

        if fallback_auth == 'enabled':
            for file in pam_radius_files:
                self._replace_radius_data(file, pam_radius_auth_fallback_disable, pam_radius_auth_fallback_enable)
            self._replace_radius_data("/etc/pam.d/pam_radius_auth", pam_radius_fallback_token_disable, pam_radius_fallback_token_enable)

        elif fallback_auth == 'disabled':
            for file in pam_radius_files:
                self._replace_radius_data(file, pam_radius_auth_fallback_enable, pam_radius_auth_fallback_disable)
            self._replace_radius_data("/etc/pam.d/pam_radius_auth", pam_radius_fallback_token_enable, pam_radius_fallback_token_disable)

        self.logger.info("Fallback authentication is enabled")

    def _apply_backup_radius_server(self, backup_rad_server):
        if not backup_rad_server:
            self.logger.info("RADIUS backup server not set !")
            return

        self.logger.info("RADIUS backup server %s is applied " % backup_rad_server)

    def _apply_backup_radius_port(self, backup_rad_port):
        if not backup_rad_port:
            self.logger.info("RADIUS backup port not set !")
            return

        self.logger.info("RADIUS backup port %s is applied" % backup_rad_port)

    def _apply_backup_radius_key(self, backup_rad_key):
        if not backup_rad_key:
            self.logger.info("RADIUS backup key not set !")
            return

        self.logger.info("RADIUS backup key %s is applied" % backup_rad_key)

    def _apply_backup_radius_timeout(self, backup_rad_timeout):
        if not backup_rad_timeout:
            self.logger.info("RADIUS backup timeout not set !")
            return

        self.logger.info("RADIUS backup timeout %s is applied" % backup_rad_timeout)

    def _apply_access_source_ip(self, ip, iptables, dst_port):
        if not ip:
            raise Exception("Access source ip address not set!")
        # create chain if not exists:
        self._create_chain(iptables)
        self._add_rule(ip, iptables, dst_port)

    def _flush_chain(self, iptables):
        cmd = '{} -F {}'.format(iptables, chain_name)
        self.execute_shell_command(cmd)

    def _remove_access_source_ip(self, iptables, chain_name):
        self._disable_cli_whitelist(iptables)
        self._flush_and_delete_chain(iptables, chain_name)

    def _apply_cli_whitelist_enabled(self, status):
        if not status:
            raise Exception("CLI Whitelist status not set!")

        if status and 'true' in str(status).lower():
            for iptables in IPTABLES:
                self._enable_cli_whitelist(iptables)
        else:
            for iptables in IPTABLES:
                self._disable_cli_whitelist(iptables)

    def _reinitialize_cli_whitelist(self, whitelist, port):
        try:
            enabled = whitelist['enabled']
        except TypeError as e:
            self.logger.debug(f"The whitelist is not enabled. Error is {e}")
            return
        for iptables in IPTABLES:
            if self._if_whitelist_enabled(iptables):
                self._disable_cli_whitelist(iptables)
            if self._if_exists_chain(iptables):
                self._flush_chain(iptables)
                self._add_deny_any_rule(iptables, port)

        if 'access-source-ip' in whitelist:
            for host in whitelist['access-source-ip']:
                if IPNetwork(host["address"]).version == 4:
                    self._apply_access_source_ip(host["address"], "/sbin/iptables", port)
                else:
                    self._apply_access_source_ip(host["address"], "/sbin/ip6tables", port)

        if enabled is True:
            for iptables in IPTABLES:
                self._enable_cli_whitelist(iptables)

    def _change_ssh_port_interfaces(self, config_ifs, old_port, new_port):
        non_management_interfaces = {key: value for key, value in config_ifs.items() if not value.get('management', False)}
        interface_names = [value['name'] for value in non_management_interfaces.values()]
        pattern = re.compile(r'(data|lte|wifi)\d+')
        filtered_interfaces = [name for name in interface_names if pattern.match(name)]

        for interface in filtered_interfaces:
            try:
                cmd = 'iptables -D INPUT -i br-{} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset'.format(
                    interface, old_port)
                output = self.execute_shell_command(cmd)
                # Check the output for any indication of errors
                if "Bad rule" in output:
                    self.logger.debug(f"Rule not found for {interface}. Skipping deletion.")
                else:
                    self.logger.debug(f"Rule deleted successfully for {interface}.")
            except Exception as e:
                self.logger.debug(f"Unexpected error: {e}")

            try:
                cmd = 'iptables -I INPUT -i br-{} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset'.format(
                    interface, new_port)
                self.execute_shell_command(cmd)
                self.logger.debug(f"Rule with {new_port} successfully added for {interface}.")
            except Exception as e:
                self.logger.debug(f"Unexpected error: {e}")

    def _reset_ip_pass_rules(self, config_ifs, ssh_port):
        try:
            vpp_if = vppif.VppApiInterface('ippass_add_del_management_entries')

            for interface in config_ifs.values():
                if interface.get('ip-passthrough', False):
                    ssh_port = int(ssh_port)
                    ip_addr = interface.get('ipv4-prefix', None)
                    if ip_addr:
                        cleaned_ip = ipaddress.ip_address(ip_addr.split('/')[0])
                        if_name, if_idx = vpp_if.search_vppif_by_hostifname(interface['name'])
                        if if_name is None or if_idx is None:
                            continue

                        host_idx = vpp_if.ippass_get_host_if_idx(if_idx)
                        if host_idx is None:
                            continue

                        ret = vpp_if.ippass_check_status(host_idx=host_idx)
                        if ret.retval != 0:
                            continue

                        add = 1
                        if vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp',
                                                                   dst_port=ssh_port, ip=cleaned_ip, add=add) != 0:
                            self.logger.error('IPPass - failed to apply management rules')
                        else:
                            self.logger.info('IPPass - {} permit CLI entry'.format('Added' if add else 'Deleted'))

        except Exception as e:
            self.logger.error(f"An error occurred: {str(e)}")
        finally:
            del vpp_if  # Ensure cleanup even if an exception occurs

    def _apply_ssh_port(self, new_port):

        try:
            if not new_port:
                self.logger.info("SSH port not set!")
                return

            # Check if the new port is the same as the current one
            result = subprocess.check_output(['grep', '-oE', '^Port[[:space:]]+[0-9]+', '/scripts/cli/sshd_config'],
                                             universal_newlines=True)
            current_port = subprocess.check_output(['awk', '{print $2}'], input=result, universal_newlines=True).strip()

            if int(new_port) == int(current_port):
                self.logger.info(f"SSH port is already set to {new_port}. No change needed.")
                return

            # Update the SSH port in the sshd_config file
            subprocess.check_call(['sed', '-i', f's/^Port .*/Port {new_port}/', '/scripts/cli/sshd_config'])
            self.logger.info(f"SSH port updated to {new_port}.")

            # Save the new port to the file (if needed)
            with open(ssh_port_file, 'w') as file:
                file.write(str(new_port))

            try:
                self._restart_ssh_service()
                self.logger.info(f"ssh.service has been restarted")
            except Exception as e:
                self.logger.error(f"The ssh.service can not be restarted due to {e}")

        except subprocess.CalledProcessError as e:
            print(f"Error executing the command: {e}")
        except Exception as e:
            self.logger.error(f"An unexpected error occurred: {e}")

    def _apply_startup_config(self):
        self.logger.info("Apply security startup configuration")
        auth = None

        # get startup config
        s_cfg = self.get_startup_config(full_path=False)

        # Security authentication section
        auth = s_cfg.get('authentication', None)
        if auth:
            auth_type = auth['auth-type']
            if auth_type == 'tacacs':
                self._apply_tacacs_server(auth['tacacs-server'])
                self._apply_tacacs_key(auth['tacacs-key'])
                self._apply_auth_type(auth_type)
            elif auth_type == 'radius':
                self._apply_radius_server(auth['radius-server'])
                self._apply_radius_port(auth['radius-port'])
                self._apply_radius_key(auth['radius-key'])
                self._apply_radius_key(auth['radius-timeout'])
                self._apply_fallback_auth(auth['fallback-auth'])
                if auth['radius-server']:
                    self._write_to_file(auth['radius-server'], auth['radius-port'], auth['radius-key'], auth['radius-timeout'])

                if auth['radius-server']:
                    self._apply_backup_radius_server(auth['backup-radius-server'])
                    self._apply_backup_radius_port(auth['backup-radius-port'])
                    self._apply_backup_radius_key(auth['backup-radius-key'])
                    self._apply_backup_radius_key(auth['backup-radius-timeout'])
                    self._write_to_file_backup(auth['backup-radius-server'], auth['backup-radius-port'], auth['backup-radius-key'], auth['backup-radius-timeout'])

                self._apply_auth_type(auth_type)
            elif auth_type == 'local':
                self._apply_auth_type(auth_type)

        wlist = s_cfg.get('cli-whitelist', None)
        if wlist:
            self._reinitialize_cli_whitelist(wlist, dst_port)

        ssh_server = s_cfg.get('ssh-server', None)
        if ssh_server:
            wlist = s_cfg.get('cli-whitelist', None)
            if wlist['enabled']:
                self._reinitialize_cli_whitelist(wlist, dst_port)
            c_ifs = self.get_interfaces()
            self._change_ssh_port_interfaces(c_ifs, dst_port, ssh_server['listen-port'])
            self._apply_ssh_port(ssh_server['listen-port'])
            if any(iface.get('ip-passthrough', False) for iface in c_ifs.values()):
                self._reset_ip_pass_rules(c_ifs, ssh_server['listen-port'])

        self.logger.info("Security startup configuration is applied")
